/*
    libtree : C++ lib that helps to work with tree-like data structures
    Copyright (C) 2019-2024 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef     ltree_treeH
#define     ltree_treeH

#include    "tree_constructor.h"
#include    "selected_data.h"
#include    "selected_values_of_additional_branches.h"

#include    <set>

#define     SUBTREE_MEM_STEP    8
#define     ENTRY_MEM_STEP      8
#define     SELDATA_MEM_STEP    1024

namespace tree{

void PopLastElOfSubtree(vector<TNode> &subtree);

//-----------------------------------------------------------------------------
// Tree of nodes

class TTree{

    protected:

        map<int, int> cells_path; // A path of cells made to get to a current cell
        map<int, set<int>> cells_connections; // Sets of intermediate cells connected with cells-leafs
        int cell = -1;   // Number of a last new cell
        int level0 = -1; // Previous level of a JSON document that is being parsed

        vector<TNode> nodes;         // Nodes of the tree
        TSelectedData selected_data; // Data, selected from DOM of the parsed XML-file
        vector<TBranch> branches;    // Branches of the tree (every branch is a mass of ids of sequented nodes)
        map<int, int> selection_index; // Count of found elements for each branch
        int n_branch_drafts = 0;     // Number of initial branch drafts
        bool first_entry_update = true;// Flag that indicates whether a current entry is been updated at least once or not

        #ifdef LTREE_DEBUGGING
        int entry_counter = 0; // Entries counter that works during a selection of data
        #endif

        // Get a number that shows that this tree is for XML or for JSON document
        virtual int Type(void) const=0;

        // Find a node of this tree using given info
        TNode* FindNode(const string &node_name, const int parent_id);

        // Data selection/extraction
        void AddItemsOfAdditionalBranches(
                const map<int, vector<string>> &values_of_additional_branches,
                const int master_branch, TSelectedDataEntry &entry,
                const size_t position_of_new_item);
        bool AddEntry(const int branches_q, TSelectedDataEntry &entry);
        bool PushNodeToSubTree(vector<TNode> &subtree, const string &node);
        pair<vector<TBranch>, bool> Intersections(
            const vector<TNode> &subtree) const;

        const int CurrentCell(const int level) const;
        void UpdateCell(const int level);
        void UpdateCellsConnections(const int current_cell);

        // Update a selection index for specified branch
        int UpdateIndex(const int master_branch);

        void FinaliseCurrentPass(const int branches_q,
                TSelectedValues &selected_values, TSelectedDataEntry &entry);

        // Special part of Select function
        virtual bool SpecialPartOFSelect(const string &file_path,
                vector<TNode> &subtree,
                TSelectedValues &selected_values,
                TSelectedDataEntry &entry,
                string &err)=0;

        void ClearBeforeNextPass(void);

        #ifdef LTREE_DEBUGGING
        // Print or produce string representation of various info that may help
        // to debug the library
        string CellsPathStr(const int level) const;
        void PrintCellsConnections(void) const;
        void PrintNewStage(const int level) const;
        void PrintMatches(const vector<TBranch> &matched_branches,
                const bool complete_match) const;
        void PrintEntryUpdate(const int branch_mark,
                const TSelectedDataEntry &entry, const unsigned indent=1);
        void PrintAdditionalUpdate(const int level, const int master_branch,
                const int branch_mark, const string &selected_value) const;
        #endif

    public:

        TSelectedData& SelectedData(void){
            return selected_data;
            }
        // Construct a tree using given base nodes and a branch draft
        bool Construct(const string &root_name,
                const vector<TBranchDraft> &branch_drafts);
        // Extract data from those nodes of the document that suit this tree
        bool Select(const string &file_path, string &err);

        TTree(void){}
        TTree(const bool v){}
        virtual ~TTree(void){}
    };

//-----------------------------------------------------------------------------

} // namespace tree

#endif
